from internal.benchmark import *
from internal.utility import *
from internal import storm
from collections import Counter
import os

# Gather execution data
def gather_exec_data(settings, logdirs):
    exec_data = OrderedDict() # Tool -> Config -> Benchmark -> Data
    benchmark_ids = []
    tools_configs = OrderedDict() # Tool X Config -> Num Supported
    for logdir_input in logdirs:
        logdir = os.path.expanduser(logdir_input)
        if not os.path.isdir(logdir):
            print("Error: Directory '{}' does not exist.".format(logdir))
    
        for res_json in [ load_json(os.path.join(logdir, f)) for f in os.listdir(logdir) if f.endswith(".json") and os.path.isfile(os.path.join(logdir, f)) ]:
            tool = res_json["tool"]
            config = res_json["configuration-id"]
            benchmark = res_json["benchmark-id"]
            exec_data.setdefault(tool, OrderedDict())
            exec_data[tool].setdefault(config, OrderedDict())
            if benchmark in exec_data[tool][config]:
                print("Error: Multiple result files found for {}.{}.{}".format(tool,config,benchmark))
            res_json["log"] = os.path.join(logdir, res_json["log"])
            exec_data[tool][config][benchmark] = res_json
            if not benchmark in benchmark_ids:
                benchmark_ids.append(benchmark)
            tools_configs.setdefault((tool,config), 0)
            if not is_not_supported(res_json):
                tools_configs[(tool,config)] += 1
    benchmark_ids.sort()
    return exec_data, benchmark_ids, tools_configs

# Reads an execution result in json format and detects if it indicates an error due to not supporting the benchmark
def is_not_supported(result_json):
    if "execution-error" not in result_json or result_json["execution-error"] == False:
        return False
    with open(result_json["log"], 'r') as logfile:
        if "storm" in result_json["tool"].lower():
            return storm.is_not_supported(logfile.read())
    return False

# Reads an execution result in json format and detects if it indicates a memory error
def is_memout(result_json):
    if "execution-error" not in result_json or result_json["execution-error"] == False:
        return False
    with open(result_json["log"], 'r') as logfile:
        if "storm" in result_json["tool"].lower():
            return storm.is_memout(logfile.read())
    return False


# Generates a csv containing runtimes. A row corresponds to a benchmark, a column corresponds to a tool/config combination.
# The first three columns contain the benchmark id, the model type, and the original format
# The last column contains the lowest runtime of the corresponding rows.
def generate_scatter_csv(settings, exec_data, benchmark_ids, tools_configs):
    MIN_VALUE = 0.05 # runtimes will be set to max(MIN_VALUE, actual runtime)
    TO_VALUE = 4000 # timeout
    MO_VALUE = 4000 # Out of memory
    NA_VALUE = 7000 # not available
    NS_VALUE = 7000 # not supported
    INC_VALUE = 12000 # Incorrect result
    ERR_VALUE = 12000 # Unknown error
    ND_VALUE = 128000 # not displayed
    
    result = [ ["benchmark", "Type", "Orig"] + ["{}.{}".format(t,c) for (t,c) in tools_configs] + ["best"] ]
    for benchmark_id in benchmark_ids:
        benchmark = get_benchmark_from_id(settings, benchmark_id)
        row = [benchmark_id, benchmark.get_model_type().lower(), benchmark.get_original_format().lower()]
        best_value = NA_VALUE
        for (tool, config) in tools_configs:
            value = NA_VALUE
            if tool == "Storm" and config == "sound" and storm.propertymethod_is_sound(benchmark):
                # This has the same code path as the sparse engine. Do not compare invocations with the exact same code path.
                value = ND_VALUE
            elif benchmark_id in exec_data[tool][config]:
                res_json = exec_data[tool][config][benchmark_id]
                if "timeout" in res_json and res_json["timeout"] == True:
                    value = TO_VALUE
                elif "result-correct" in res_json and res_json["result-correct"] == False:
                    value = INC_VALUE
                elif "execution-error" in res_json and res_json["execution-error"] == True:
                    if is_memout(res_json):
                        value = MO_VALUE
                    elif is_not_supported(res_json):
                        value = NS_VALUE
                    else:
                        print("Unexpected error for '{}.{}.{}'".format(tool, config, benchmark_id))
                        value = ERR_VALUE
                elif "timeout" in res_json and "wallclock-time" in res_json:
                    value = max(MIN_VALUE,float(res_json["wallclock-time"]))
                    best_value = min(best_value, value)
                else:
                    print("Unexpected execution result for '{}.{}.{}'".format(tool, config, benchmark_id))
            row.append(str(value))
        row.append(str(best_value))
        result.append(row)
    return result


# Generates a csv containing runtimes. The first column denotes the row indices. Each of the remaining column corresponds to a tool/config combination. The last column corresponds to the fastest tool/config
# An entry in the ith row corresponds to the runtime of the ith fastest benchmark
def generate_quantile_csv(settings, exec_data, benchmark_ids, tools_configs):
    MIN_VALUE = 1.0 # runtimes will be set to max(MIN_VALUE, actual runtime)
    runtimes_best_dict = OrderedDict()
    for (t,c) in tools_configs:
        runtimes_best_dict[t] = OrderedDict()
    result = [ ["n"] + ["{}.{}shifted".format(t,c) for (t,c) in tools_configs]  + ["{}.bestshifted".format(tool) for tool in runtimes_best_dict]] # append 'shifted' for compatibility with qcomp latex
    runtimes = OrderedDict()
    for (tool, config) in tools_configs:
        runtimes_tc = []
        for benchmark_id in benchmark_ids:
            if benchmark_id in exec_data[tool][config]:
                res_json = exec_data[tool][config][benchmark_id]
                if "timeout" in res_json and res_json["timeout"] == False and ("result-correct" not in res_json or res_json["result-correct"] == True) and ("execution-error" not in res_json or res_json["execution-error"] == False) and "wallclock-time" in res_json:
                    runtimes_tc.append(max(MIN_VALUE,float(res_json["wallclock-time"])))
                    if benchmark_id not in runtimes_best_dict[tool]:
                        runtimes_best_dict[tool][benchmark_id] = runtimes_tc[-1]
                    else:
                        runtimes_best_dict[tool][benchmark_id] = min(runtimes_best_dict[tool][benchmark_id], runtimes_tc[-1])
        runtimes_tc.sort()
        runtimes["{}.{}".format(tool,config)] = runtimes_tc
    for tool in runtimes_best_dict:
        runtimes_best = [ runtimes_best_dict[tool][b] for b in runtimes_best_dict[tool] ]
        runtimes_best.sort()
        runtimes["{}.best".format(tool)] = runtimes_best
    for i in range(len(benchmark_ids)):
        row = [str(i+1)]
        for tc in runtimes:
            if i < len(runtimes[tc]):
                row.append(str(runtimes[tc][i]))
            else:
                row.append("")
        result.append(row)
    print("Correctly solved benchmarks (of {}):".format(len(benchmark_ids)))
    for tc in runtimes:
        print("\t{}: {}".format(tc, len(runtimes[tc])))
    return result

def generate_summary_table(settings, exec_data, benchmark_ids, tools_configs):
    solved=Counter()
    solved_fastest1=Counter()
    solved_fastest2=Counter()
    not_supported=Counter()
    timeout=Counter()
    memout=Counter()
    incorrect=Counter()
    error=Counter()
    for benchmark_id in benchmark_ids:
        benchmark = get_benchmark_from_id(settings, benchmark_id)
        best_value = 100000
        solving_times = dict()
        for (tool, config) in tools_configs:
            if benchmark_id in exec_data[tool][config]:
                res_json = exec_data[tool][config][benchmark_id]
                if "timeout" in res_json and res_json["timeout"] == True:
                    timeout[(tool,config)] += 1
                elif "result-correct" in res_json and res_json["result-correct"] == False:
                    incorrect[(tool,config)] += 1
                elif "execution-error" in res_json and res_json["execution-error"] == True:
                    if is_memout(res_json):
                        memout[(tool,config)] += 1
                    elif is_not_supported(res_json):
                        not_supported[(tool,config)] += 1
                    else:
                        print("Unexpected error for '{}.{}.{}'".format(tool, config, benchmark_id))
                        error[(tool,config)] += 1
                elif "timeout" in res_json and "wallclock-time" in res_json:
                    solved[(tool,config)] += 1
                    value = float(res_json["wallclock-time"])
                    solving_times[(tool,config)] = value
                    if value < best_value:
                        best_value = value
                else:
                    print("Unexpected execution result for '{}.{}.{}'".format(tool, config, benchmark_id))
            else:
                not_supported[(tool,config)] += 1
        # count fastest
        for tc in solving_times:
            if solving_times[tc] <= 1.01 * best_value:
                solved_fastest1[tc] += 1
            if solving_times[tc] <= 1.5 * best_value:
                solved_fastest2[tc] += 1
        solved["best"] += min(len(solving_times),1)

    result = r"""\begin{tabular}{r|""" + "r" * len(tools_configs) + """}""" + "\n"
    for (t,c) in tools_configs:
        result += "\t& " +   """\multicolumn{1}{c}{\engine{""" + c + "}}\n"
    result += "\\\\\\hline\n"
    result += """\#solved     &""" + "\t&".join([" {}".format(solved[tc]) for tc in tools_configs]) + "\\\\\n"
    result += """\#not supp.  &""" + "\t&".join([" {}".format(not_supported[tc]) for tc in tools_configs]) + "\\\\\n"
    result += """\#time-outs  &""" + "\t&".join([" {}".format(timeout[tc]) for tc in tools_configs]) + "\\\\\n"
    result += """\#mem-outs   &""" + "\t&".join([" {}".format(memout[tc]) for tc in tools_configs]) + "\\\\\n"
    result += """\#incorrect  &""" + "\t&".join([" {}".format(incorrect[tc]) for tc in tools_configs]) + "\\\\\n"
    result += """\#error      &""" + "\t&".join([" {}".format(error[tc]) for tc in tools_configs]) + "\\\\\n"
    result += """\#fastest101 &""" + "\t&".join([" {}".format(solved_fastest1[tc]) for tc in tools_configs]) + "\\\\\n"
    result += """\#fastest150 &""" + "\t&".join([" {}".format(solved_fastest2[tc]) for tc in tools_configs]) + "\\\\\n"
    result += r"""\end{tabular}""" + "\n"
    result += "% A total of {} instances could be solved.".format(solved["best"])
    return result 
     
    

# Aux function for writing in files with proper indention
def write_line(file, indention, content):
    file.write("\t"*indention + content + "\n")

# Generates an html log page for the given result within output_dir/logs/
def create_log_page(settings, result_json, output_dir):
    b = get_benchmark_from_id(settings, result_json["benchmark-id"])
    if not "log" in result_json:
        raise AssertionError("Expected a log file.")
    with open(result_json["log"], 'r') as logfile:
        logs = logfile.read().split("#" * 40)
    f_path = os.path.join("logs/", os.path.basename(result_json["log"])[:-4] + ".html")
    with open(os.path.join(output_dir, f_path), 'w') as f:
        indention = 0
        write_line(f, indention, "<!DOCTYPE html>")
        write_line(f, indention, "<html>")
        write_line(f, indention, "<head>")
        indention += 1
        write_line(f, indention, '<meta charset="UTF-8">')
        write_line(f, indention, "<title>{}.{} - {} {} {}</title>".format(result_json["tool"], result_json["configuration-id"], b.get_model_short_name(), b.get_property_name(), b.get_parameter_values_string()))
#        write_line(f, indention, '<link rel="stylesheet" type="text/css" href="{}">'.format("http://qcomp.org/style.css")) # TODO: Might want to add another style
        write_line(f, indention, '<link rel="stylesheet" type="text/css" href=../style.css>') # TODO: Might want to add another style
        indention -= 1
        write_line(f, indention, '</head>')
        write_line(f, indention, '<body>')
        write_line(f, indention, '<h1>{}.{}</h1>'.format(result_json["tool"],result_json["configuration-id"]))

        write_line(f, indention, '<div class="box">')
        indention += 1
        write_line(f, indention, '<div class="boxlabelo"><div class="boxlabelc">Benchmark</div></div>')
        write_line(f, indention, '<table style="margin-bottom: 0.75ex;">')
        indention += 1
        write_line(f, indention, '<tr><td>Model:</td><td><a href="{}">{}</a> <span class="tt">v.{}</span> ({})</td></tr>'.format("http://qcomp.org/benchmarks/index.html#{}".format(b.get_model_short_name()), b.get_model_short_name(), b.index_json["version"], b.get_model_type().upper()))
        write_line(f, indention, '<tr><td>Parameter(s)</td><td>{}</td></tr>'.format(", ".join(['<span class="tt">{}</span> = {}'.format(p["name"], p["value"]) for p in b.get_parameters()])))
        write_line(f, indention, '<tr><td>Property:</td><td><span class="tt">{}</span> ({})</td></tr>'.format(b.get_property_name(), b.get_property_type()))
        indention -= 1
        write_line(f, indention, "</table>")
        indention -= 1
        write_line(f, indention, "</div>")

        write_line(f, indention, '<div class="box">')
        indention += 1
        write_line(f, indention, '<div class="boxlabelo"><div class="boxlabelc">Invocation ({})</div></div>'.format(result_json["configuration-id"]))
        f.write('\t' * indention + '<pre style="overflow: auto; padding-bottom: 1.5ex; padding-top: 1ex; font-size: 15px; margin-bottom: 0ex;  margin-top: 0ex;">')
        commands_str = "\n".join(result_json["commands"])
        for filtered_path in settings.filtered_paths():
            commands_str = commands_str.replace(filtered_path, "")
        f.write(commands_str)
        f.write('</pre>\n')
        write_line(f, indention, result_json["invocation-note"])
        indention -= 1
        write_line(f, indention, "</div>")

        write_line(f, indention, '<div class="box">')
        indention += 1
        write_line(f, indention, '<div class="boxlabelo"><div class="boxlabelc">Execution</div></div>')
        write_line(f, indention, '<table style="margin-bottom: 0.75ex;">')
        indention += 1
        if result_json["timeout"]:
            write_line(f, indention, '<tr><td>Walltime:</td><td style="color: red;">&gt {}s (Timeout)</td></tr>'.format(result_json["time-limit"]))
        else:
            write_line(f, indention, '<tr><td>Walltime:</td><td style="tt">{}s</td></tr>'.format(result_json["wallclock-time"]))
            return_codes = []
            if "return-codes" in result_json:
                return_codes = result_json["return-codes"]
            if result_json["execution-error"]:
                write_line(f, indention, '<tr><td>Return code:</td><td style="tt; color: red;">{}</td></tr>'.format(", ".join([str(rc) for rc in return_codes])))
            else:
                write_line(f, indention, '<tr><td>Return code:</td><td style="tt">{}</td></tr>'.format(", ".join([str(rc) for rc in return_codes])))
        first = True
        for note in result_json["notes"]:
            write_line(f, indention, '<tr><td>{}</td><td>{}</td></tr>'.format("Note(s):" if first else "", note))
            first = False
        if "relative-error" in result_json:
            write_line(f, indention, '<tr><td>Relative Error:</td><td style="tt{}">{}</td></tr>'.format("" if result_json["result-correct"] else "; color: red", result_json["relative-error"]))
        indention -= 1
        write_line(f, indention, "</table>")
        indention -= 1
        write_line(f, indention, "</div>")

        for log in logs:
            for filtered_path in settings.filtered_paths():
                log = log.replace(filtered_path, "")
            pos = log.find("\n", log.find("Output:\n")) + 1
            pos_end = log.find("#############################", pos)
            if pos_end < 0:
                pos_end = len(log)
            log_str = log[pos:pos_end].strip()
            if len(log_str) != 0:
                write_line(f, indention, '<div class="box">')
                indention += 1
                write_line(f, indention, '<div class="boxlabelo"><div class="boxlabelc">Log</div></div>')
                f.write("\t" * indention + '<pre style="overflow:auto; padding-bottom: 1.5ex">')
                f.write(log_str)
                write_line(f, indention, '</pre>')
                indention -= 1
                write_line(f, indention, "</div>")

            pos = log.find("##############################Output to stderr##############################\n")
            if pos >= 0:
                pos = log.find("\n", pos) + 1
                write_line(f, indention, '<div class="box">')
                indention += 1
                write_line(f, indention, '<div class="boxlabelo"><div class="boxlabelc">STDERR</div></div>')
                f.write("\t" * indention + '<pre style="overflow:auto; padding-bottom: 1.5ex">')
                pos_end = log.find("#############################", pos)
                if pos_end < 0:
                    pos_end = len(log)
                f.write(log[pos:pos_end].strip())
                write_line(f, indention, '</pre>')
                indention -= 1
                write_line(f, indention, "</div>")
        write_line(f, indention, "</body>")
        write_line(f, indention, "</html>")
    return f_path

# Generates an interactive html table from the results
def generate_table(settings, exec_data, benchmark_ids, tools_configs, output_dir):
    SHOW_UNSUPPORTED = True # Also add entries for benchmarks that are known to be unsupported
    
    ensure_directory(output_dir)
    ensure_directory(os.path.join(output_dir, "logs/"))
    
    first_tool_col = 6
    num_cols = first_tool_col + len(tools_configs)

    with open (os.path.join(output_dir, "table.html"), 'w') as tablefile:
        tablefile.write(r"""<!DOCTYPE html>
    <html>
    <head>
      <meta charset="UTF-8">
      <title>Benchmark results</title>
      <link rel="stylesheet" type="text/css" href="style.css">
      <link rel="stylesheet" type="text/css" href="https://cdn.datatables.net/1.10.13/css/jquery.dataTables.min.css">
      <link rel="stylesheet" type="text/css" href="https://cdn.datatables.net/buttons/1.2.4/css/buttons.dataTables.min.css">
      <link rel="stylesheet" type="text/css" href="https://cdn.datatables.net/fixedheader/3.1.2/css/fixedHeader.dataTables.min.css">

      <script type="text/javascript" language="javascript" charset="utf8" src="https://code.jquery.com/jquery-1.12.4.js"></script>
      <script type="text/javascript" language="javascript" charset="utf8" src="https://cdn.datatables.net/1.10.13/js/jquery.dataTables.min.js"></script>
      <script type="text/javascript" language="javascript" charset="utf8" src="https://cdn.datatables.net/fixedheader/3.1.2/js/dataTables.fixedHeader.min.js"></script>
      <script type="text/javascript" language="javascript" charset="utf8" src="https://cdn.datatables.net/buttons/1.2.4/js/dataTables.buttons.min.js"></script>
      <script type="text/javascript" language="javascript" charset="utf8" src="https://cdn.datatables.net/buttons/1.2.4/js/buttons.colVis.min.js"></script>

      <script>
        $(document).ready(function() {
          // Set correct file
          $("#content").load("data.html");
        } );

        function updateBest(table) {
          // Remove old best ones
          table.cells().every( function() {
            $(this.node()).removeClass("best");
          });
          table.rows().every( function ( rowIdx, tableLoop, rowLoop ) {
              var bestValue = -1
              var bestIndex = -1
              $.each( this.data(), function( index, value ){
                if (index > 5 && table.column(index).visible()) {
    			    var text = $(value).text()
    	            if (["TO", "ERR", "INC", "MO", "NS", ""].indexOf(text) < 0) {
    				    var number = parseFloat(text);
    	                if (bestValue == -1 || bestValue > number) {
    	                  // New best value
    	                  bestValue = number;
    	                  bestIndex = index;
    	                }
    				  }
    			  }
              });
              // Set new best
              if (bestIndex >= 0) {
                $(table.cell(rowIdx, bestIndex).node()).addClass("best");
              }
          } );
      }
      </script>
    </head>
    """)
        indention = 0
        write_line(tablefile, indention, "<body>")
        write_line(tablefile, indention, "<div>")
        indention +=1
        write_line(tablefile, indention, '<table id="table" class="display">')
        indention += 1
        write_line(tablefile, indention, '<thead>')
        indention += 1
        write_line(tablefile, indention, '<tr>')
        indention += 1
        for head in ["Model", "Type", "Original", "Parameters", "Property", "Type"] + ["{}.{}".format(t,c) for (t,c) in tools_configs]:
            write_line(tablefile, indention, '<th>{}</th>'.format(head))
        indention -= 1
        write_line(tablefile, indention, '</tr>')
        indention -= 1
        write_line(tablefile, indention, '</thead>')
        write_line(tablefile, indention, '<tbody>')
        indention += 1

        for benchmark_id in benchmark_ids:
            b = get_benchmark_from_id(settings, benchmark_id)
            write_line(tablefile, indention, '<tr>')
            indention += 1
            write_line(tablefile, indention, '<td><a href="{}">{}</a></td>'.format("http://qcomp.org/benchmarks/index.html#{}".format(b.get_model_short_name()), b.get_model_short_name()))
            write_line(tablefile, indention, '<td>{}</td>'.format(b.get_model_type().upper()))
            write_line(tablefile, indention, '<td>{}</td>'.format(b.get_original_format()))
            write_line(tablefile, indention, '<td>{}</td>'.format(b.get_parameter_values_string()))
            write_line(tablefile, indention, '<td>{}</td>'.format(b.get_property_name()))
            write_line(tablefile, indention, '<td>{}</td>'.format(b.get_short_property_type()))
            for (t,c) in tools_configs:
                cell_content = ""
                if benchmark_id in exec_data[t][c]:
                    res_json = exec_data[t][c][benchmark_id]
                    link_attributes = ""
                    if "timeout" in res_json and res_json["timeout"] == True:
                        res_str = "TO"
                        link_attributes = " class='timeout'"
                    elif "result-correct" in res_json and res_json["result-correct"] == False:
                        res_str = "INC"
                        link_attributes = " class='incorrect'"
                    elif "execution-error" in res_json and res_json["execution-error"] == True:
                        if is_memout(res_json):
                            res_str = "MO"
                            link_attributes = " class='memout'"
                        elif is_not_supported(res_json):
                            res_str = "NS" if SHOW_UNSUPPORTED else None
                            link_attributes = " class='unsupported'"
                        else:
                            res_str = "ERR"
                            link_attributes = " class='error'"
                            print("Unexpected error for '{}.{}.{}' (found while creating table)".format(t, c, benchmark_id))
                    elif "timeout" in res_json and "execution-error" in res_json and "wallclock-time" in res_json:
                        res_str = "%.1f" % res_json["wallclock-time"]
                    else:
                        print("Unexpected execution result for '{}.{}.{}'".format(t, c, benchmark_id))
                    if res_str is not None:
                        logpage = create_log_page(settings, res_json, output_dir)
                        cell_content = "<a href='{}' {}>{}</a>".format(logpage, link_attributes, res_str)                
                write_line(tablefile, indention, '<td>{}</td>'.format(cell_content))
            indention -= 1
            write_line(tablefile, indention, '</tr>')
        indention -= 1
        write_line(tablefile, indention, '</tbody>')
        indention -= 1
        indention -= 1
        write_line(tablefile, indention, '</table>')
        write_line(tablefile, indention, "<script>")
        indention +=1
        write_line(tablefile, indention, 'var table = $("#table").DataTable( {')
        indention += 1
        write_line(tablefile, indention, '"paging": false,')
        write_line(tablefile, indention, '"autoWidth": false,')
        write_line(tablefile, indention, '"info": false,')
        write_line(tablefile, indention, 'fixedHeader: {')
        indention += 1
        write_line(tablefile, indention, '"header": true,')
        indention -= 1
        write_line(tablefile, indention, '},')
        write_line(tablefile, indention, '"dom": "Bfrtip",')
        write_line(tablefile, indention, 'buttons: [')
        indention += 1
        for columnIndex in range(first_tool_col, num_cols):
            write_line(tablefile, indention, '{')
            indention += 1
            write_line(tablefile, indention, 'extend: "columnsToggle",')
            write_line(tablefile, indention, 'columns: [{}],'.format(columnIndex))
            indention -= 1
            write_line(tablefile, indention, "},")
        tool_columns = [i for i in range(first_tool_col, num_cols)]
        for text, show, hide in zip(["Show all", "Hide all"], [tool_columns, []], [[], tool_columns]):
            write_line(tablefile, indention, '{')
            indention += 1
            write_line(tablefile, indention, 'extend: "colvisGroup",')
            write_line(tablefile, indention, 'text: "{}",'.format(text))
            write_line(tablefile, indention, 'show: {},'.format(show))
            write_line(tablefile, indention, 'hide: {}'.format(hide))
            indention -= 1
            write_line(tablefile, indention, "},")
        indention -= 1
        write_line(tablefile, indention, "],")
        indention -= 1
        write_line(tablefile, indention, "});")
        indention -= 1
        write_line(tablefile, indention, "")
        indention += 1
        write_line(tablefile, indention, 'table.on("column-sizing.dt", function (e, settings) {')
        indention += 1
        write_line(tablefile, indention, "updateBest(table);")
        indention -= 1
        write_line(tablefile, indention, "} );")
        indention -= 1
        write_line(tablefile, indention, "")
        indention += 1
        write_line(tablefile, indention, "updateBest(table);")
        indention -= 1
        write_line(tablefile, indention, "</script>")
        indention -= 1
        write_line(tablefile, indention, "</div>")
        write_line(tablefile, indention, "</body>")
        write_line(tablefile, indention, "</html>")

    with open (os.path.join(output_dir, "style.css"), 'w') as stylefile:
#        write_line(stylefile, 0, '@import url("{}");'.format(os.path.join(qcomp_root, "fonts/Tajawal/Tajawal.css"))) #TODO
        stylefile.write(r"""

    .best {
        background-color: lightgreen;
    }
    .error {
    	font-weight: bold;
    	background-color: lightcoral;
    }
    .incorrect {
        background-color: orange;
    	font-weight: bold;
    }
    .timeout {
        background-color: lightgray;
    }
    .memout {
        background-color: lightgray;
    }
    .unsupported {
        background-color: yellow;
    }

    h1 {
    	font-size: 28px; font-weight: bold;
    	color: #000000;
    	padding: 1px; margin-top: 20px; margin-bottom: 1ex;
    }

    tt, .tt {
    	font-family: 'Courier New', monospace; line-height: 1.3;
    }

    .box {
    	margin: 2.5ex 0ex 1ex 0ex; border: 1px solid #D0D0D0; padding: 1.6ex 1.5ex 1ex 1.5ex; position: relative;
    }

    .boxlabelo {
    	position: absolute; pointer-events: none; margin-bottom: 0.5ex;
    }

    .boxlabel {
    	position: relative; top: -3.35ex; left: -0.5ex; padding: 0px 0.5ex; background-color: #FFFFFF; display: inline-block;
    }
    .boxlabelc {
    	position: relative; top: -3.17ex; left: -0.5ex; padding: 0px 0.5ex; background-color: #FFFFFF; display: inline-block;
    }
    """)
