from .benchmark import Benchmark
from .invocation import Invocation
from .execution import *
from .configuration import *

def get_name():
    """ should return the name of the tool as listed on http://qcomp.org/competition/2019/"""
    return "Storm"

def test_installation(settings, configuration = None):
    """
    Performs a quick check to test wether the installation works. 
    Returns an error message if something went wrong and 'None' otherwise.
    """
    is_dft = configuration is not None and "dft" in configuration.identifier
    if is_dft:
        storm_executable = os.path.join(settings.storm_binary_dir(), "storm-dft")    
    else:
        storm_executable = os.path.join(settings.storm_binary_dir(), "storm")    
    if not os.path.exists(storm_executable):
         return "Binary '{}' does not exist.".format(storm_executable)    
    command_line = storm_executable + " {}".format("" if configuration is None else configuration.command)
    try:
        test_out, test_time, test_code = execute_command_line(command_line, 10)
        if test_code != 0:
            return "Error while executing:\n\t{}\nNon-zero return code: '{}'.".format(command_line, test_code)
    except KeyboardInterrupt:
        return "Error: Execution interrupted."
    except Exception:
        return "Error while executing\n\t{}\n".format(command_line)


def propertymethod_is_sound(benchmark : Benchmark):
    """ Auxiliary function that returns true if the default method used for this property is sound."""
    # step and time bounded properties are always sound
    # This does not check for properties with only a lower time/step/reward bound.
    # However such properties are not in QComp2019
    if "step-bounded" in benchmark.get_property_type():
        return True
    if "time-bounded" in benchmark.get_property_type():
        return True
    return False    
    
    
def is_benchmark_supported(benchmark : Benchmark, configuration : Configuration):
    """ Auxiliary function that returns True if the provided benchmark is not supported by Storm and no known external conversion tool can help."""
    # Storm does not support CTMCs with infinite state-spaces
    if benchmark.is_prism_inf() and benchmark.is_ctmc():
        return False
        
    return True


def get_invocation(settings, benchmark : Benchmark, configuration : Configuration):
    """
    Returns an invocations that invokes the tool for the given benchmark and the given storm configuration.
    It can be assumed that the current directory is the directory from which execute_invocations.py is executed.
    For QCOMP 2019, this should return a list of size at most two, where
    the first entry (if present) corresponds to the default configuration of the tool and
    the second entry (if present) corresponds to an optimized setting (e.g., the fastest engine and/or solution technique for this benchmark).
    Please only provide two invocations if there is actually a difference between them.
    If this benchmark is not supported, an empty list has to be returned.
    For testing purposes, the script also allows to return more than two invocations.
    """
    general_arguments = "--timemem" # Prints some timing and memory information
    
    invocation = Invocation()
    invocation.tool = get_name()
    invocation.configuration_id = configuration.identifier
    invocation.note = configuration.note
    invocation.benchmark_id = benchmark.get_identifier()
    
    if is_benchmark_supported(benchmark, configuration):
        bdir = benchmark.get_directory()
        if "dft" in configuration.identifier:
            if benchmark.is_galileo():
                storm_dft_executable = os.path.join(settings.storm_binary_dir(), "storm-dft")
                storm_dft_command = "--dftfile " + os.path.join(bdir, benchmark.get_galileo_filename())  + " "
                
                prop = benchmark.get_property_jani()
                # Try to detect what the property is about. TODO: This is not very stable and could easily produce the wrong command line
                success = False
                if "expression" in prop and "values" in prop["expression"] and "op" in prop["expression"]["values"] and "exp" in prop["expression"]["values"]:
                    op = prop["expression"]["values"]["op"]                        
                    exp = prop["expression"]["values"]["exp"]
                    if isinstance(exp, OrderedDict) and "time-bounds" in exp and "upper" in exp["time-bounds"] and "lower" not in exp["time-bounds"]:
                        time_bound = exp["time-bounds"]["upper"]
                        for p in benchmark.get_parameters():
                            if p["name"] == time_bound:
                                time_bound = p["value"]
                                break
                        storm_dft_command += "--timebound {} ".format(time_bound)
                        if op in ["Pmax", "Pmin"]:
                            storm_dft_command += "--" + op[1:] + " "
                            success = True
                    if op in ["Emin", "Emax"]:
                        storm_dft_command += "--expectedtime --" + op[1:] + " "
                        success = True
                if success:
                    invocation.add_command(storm_dft_executable + " " + storm_dft_command + " " + general_arguments)
                else:
                    invocation.note += " Not supported: Unable to detect dft property."
            else:
                invocation.note += " Not supported: storm-dft only supports dynamic fault trees."
        else:        
            storm_executable = os.path.join(settings.storm_binary_dir(), "storm")    
    
            if (benchmark.is_prism() or benchmark.is_prism_ma()) and not benchmark.is_pta():
                benchmark_arguments = "--prism {} --prop {} {}".format(os.path.join(bdir, benchmark.get_prism_program_filename()), os.path.join(bdir, benchmark.get_prism_property_filename()), benchmark.get_property_name())
                if benchmark.get_open_parameter_def_string() != "":
                    benchmark_arguments += " --constants {}".format(benchmark.get_open_parameter_def_string())
                if benchmark.is_ctmc():
                    benchmark_arguments += " --prismcompat"
                    invocation.note += " Use `--prismcompat` to ensure compatibility with prism benchmark."
            else:
                # For jani input, it might be the case that preprocessing is necessary using moconv
                janifile = benchmark.get_janifilename()
                par_defs = benchmark.get_open_parameter_def_string()
                moconv_options = []
                if "nondet-selection" in benchmark.get_jani_features():
                    moconv_options.append("--remove-disc-nondet")
                    invocation.note += " Use moconv to handle currently unsupported jani feature 'nondet-selection'."
                if benchmark.is_pta():
                    moconv_options.append("--digital-clocks")
                    if benchmark.load_jani_file()["type"] == "sta":
                        moconv_options.append(" --unroll-distrs")
                    invocation.note += " Use moconv to convert the PTA to an MDP using digital-clocks semantics."
                if len(moconv_options) != 0:    
                    if par_defs != "":
                        moconv_options.append(" --experiment " + par_defs)
                        par_defs = ""
                    if not os.path.isfile(os.path.join(bdir, "converted_" + janifile)):
                        print("Moconv call required:\nmoconv {} {} --output {} --overwrite\n".format(os.path.join(bdir, janifile), " ".join(moconv_options), os.path.join(bdir, "converted_" + janifile)))
                    janifile = "converted_" + janifile
                benchmark_arguments = "--jani {} --janiproperty {}".format(os.path.join(bdir, janifile), benchmark.get_property_name())
                if par_defs != "":
                    benchmark_arguments += " --constants " + par_defs
            
            # If this is a sound computation, we set the precision to the required one. Otherwise, we keep the default value.
            if "--sound " in configuration.command or propertymethod_is_sound(benchmark):
                if ("--exact" not in configuration.command):
                    general_arguments += " --precision {}".format(float(settings.goal_precision()))
                    
            invocation.add_command(storm_executable + " " + benchmark_arguments + " " + configuration.command + " " + general_arguments)
    else:
        invocation.note += " Benchmark not supported by Storm."    
    return invocation



def get_result(benchmark : Benchmark, execution : Execution):
    """
    Returns the result of the given execution on the given benchmark.
    This method is called after executing the commands of the associated invocation.
    One can either find the result in the tooloutput (as done here) or
    read the result from a file that the tool has produced.
    The returned value should be either 'true', 'false', a decimal number, or a fraction.
    """
    invocation = execution.invocation
    log = execution.concatenate_logs()
    if "Storm-dft" in log:
        pos = log.find("Result: [")
        if pos >= 0:
            pos = pos + len("Result: [")
            pos_e = log.find("]\n", pos)
            if pos_e >= 0:
                return log[pos:pos_e]

    pos = log.find("Model checking property \"{}\":".format(benchmark.get_property_name()))
    if pos < 0:
        return None
    pos = log.find("Result (for initial states): ", pos)
    if pos < 0:
        return None
    pos = pos + len("Result (for initial states): ")
    eol_pos = log.find("\n", pos)
    result = log[pos:eol_pos]
    pos_appr = result.find("(approx. ")
    if pos_appr >= 0:
        result = result[:pos_appr]
    return result
    
def is_not_supported(logfile):
    """
    Returns true if the logfile contains error messages that mean that the input is not supported.
    """
    # if one of the following error messages occurs, we are sure that the model is not supported.
    known_messages = []
    known_messages.append("The model type Markov Automaton is not supported by the hybrid engine.")
    known_messages.append("The model type Markov Automaton is not supported by the dd engine.")
    known_messages.append("The model type CTMC is not supported by the dd engine.")
    known_messages.append("Cannot build symbolic model from JANI model whose system composition that refers to the automaton ")
    known_messages.append("Cannot build symbolic model from JANI model whose system composition refers to the automaton ")
    known_messages.append("The symbolic JANI model builder currently does not support assignment levels.")
    for m in known_messages:
        if m in logfile:
            return True
    
    pos_message = logfile.find("Inspections (defined in line ")
    if pos_message >= 0:
        end_of_line = logfile.find("\n",pos_message)
        if logfile[:end_of_line].endswith(") are not supported."): return True
    
    # The following only indicates unsupported if there is no other error message before
    pos_message = logfile.find("Property is unsupported by selected engine/settings.")
    if pos_message >= 0:
        end_of_line_before = logfile[:pos_message].rfind("\n")
        if "ERROR" not in logfile[:end_of_line_before]:
            start_of_line_before = logfile[:end_of_line_before].rfind("\n") + 1
            line_before = logfile[start_of_line_before:end_of_line_before]
            if line_before.startswith("Model checking property"): return True


    return False

def is_memout(logfile):
    """
    Returns true if the logfile indicates an out of memory situation.
    Assumes that a result could not be parsed successfully.
    """
    known_messages = []
    known_messages.append("Maximum memory exceeded.")
    known_messages.append("BDD Unique table full")
    for m in known_messages:
        if m in logfile:
            return True
    # if there is no error message and no result is produced, we assume out of memory.
    return "ERROR" not in logfile
