import os
import sys
import time
import math
import csv
import json
import shutil
from decimal import *
from fractions import *

from collections import OrderedDict
from .configuration import Configuration

def load_json(path : str):
    with open(path, 'r', encoding='utf-8-sig') as json_file:
        return json.load(json_file, object_pairs_hook=OrderedDict)

def save_json(json_data, path : str):
    with open(path, 'w') as json_file:
        json.dump(json_data, json_file, ensure_ascii=False, indent='\t')

def load_csv(path : str, delim='\t'):
    with open(path, 'r') as csv_file:
        return list(csv.reader(csv_file, delimiter=delim))

def save_csv(csv_data, path : str, delim=';'):
    with open(path, 'w') as csv_file:
        writer = csv.writer(csv_file, delimiter=delim)
        writer.writerows(csv_data)

def ensure_directory(path : str):
    if not os.path.exists(path):
        os.makedirs(path)

def is_valid_filename(name : str, invalid_chars = None):
    if invalid_chars is not None:
        for c in invalid_chars:
            if c in name:
                return False
    try:
        if os.path.isfile(name):
            return True
        open(name, 'a').close()
        os.remove(name)
    except IOError:
        return False
    return True

def remove_directory_contents(directory, exluded = []):
    for name in os.listdir(directory):
        if name not in exluded:
            try:
                path = os.path.join(directory, name)
                remove_file_or_dir(path)
            except Exception:
                print("Unable to remove '{}'".format(path))

def remove_file_or_dir(name):
    if os.path.isdir(name):
        shutil.rmtree(name)
    else:
        os.remove(name)

def is_bool(expr):
    if isinstance(expr, bool):
        return True
    try:
        return expr.lower() in ["true", "false"]
    except:
        return False

def is_inf(expr):
    try:
        return math.isinf(Decimal(expr))
    except Exception:
        return False

def is_number(expr):
    if is_bool(expr):
        return False
    if is_inf(expr):
        return True
    try:
        Fraction(expr)
    except Exception:
        return False
    return True

def is_interval(expr):
    try:
        if is_number(expr["lower"]) and is_number(expr["upper"]):
            return True
    except (InvalidOperation, KeyError, TypeError):
        pass
    return False

def is_number_or_interval(expr):
    return is_number(expr) or is_interval(expr)

def try_to_number(expr):
    if is_number(expr):
        if is_inf(expr):
            return Decimal(expr)
        else:
            return Fraction(expr)
    try:
        return try_to_number(expr["num"]) / try_to_number(expr["den"])
    except Exception:
        return expr

def try_to_bool_or_number(expr):
    if is_bool(expr):
        if (isinstance(expr, str)):
            if expr.lower() == "true":
                return True
            elif expr.lower() == "false":
                return False
        return bool(expr)
    return try_to_number(expr)

def get_decimal_representation(number):
    if is_number(number):
        return Decimal(number)
    else:
        return Decimal(number["num"]) / Decimal(number["den"])

def try_to_float(expr):
    # expr might be too large for float
    try:
        return float(expr)
    except Exception:
        return expr

def get_absolute_error(reference_value, result_value):
    return get_error(False, reference_value, result_value)

def get_relative_error(reference_value, result_value):
    return get_error(True, reference_value, result_value)

def get_error(relative : bool, reference_value, result_value):
    result_value = try_to_number(result_value)
    if is_interval(reference_value):
        u = try_to_number(reference_value["upper"])
        l = try_to_number(reference_value["lower"])
        if is_inf(result_value) and not is_inf(u):
            return math.inf
        elif result_value < l:
            return get_error(relative, l, result_value)
        elif result_value > u:
            return get_error(relative, u, result_value)
        else:
            return Fraction(0)
    reference_value = try_to_number(reference_value)
    if relative and reference_value == 0:
        if result_value == 0:
            return Fraction(0)
        else:
            return math.inf
    elif is_inf(reference_value) and is_inf(result_value):
        return Fraction(0)
    elif is_inf(reference_value) or is_inf(result_value):
        return math.inf

    diff = abs(reference_value - result_value)
    if relative:
        return diff / reference_value
    else:
        return diff

def is_result_correct(settings, reference, result):
    if is_number_or_interval(reference) != is_number(result):
        return False
    if is_number_or_interval(reference):
        return get_error(settings.is_relative_precision(), reference, result) <= settings.goal_precision()
    else:
        return reference == result

class Progressbar(object):
    def __init__(self, max_value, label="Progress", width=50, delay=0.5):
        self.progress = 0
        self.max_value = max_value
        self.label = label
        self.width = width
        self.delay = delay
        self.last_time_printed = time.time()
        sys.stdout.write("\n")
        self.print_progress(0)

    def print_progress(self, value):
        now = time.time()
        if now - self.last_time_printed >= self.delay or value == self.max_value or value == 0:
            if (self.max_value == 0):
                progress = self.width
            else:
                progress = (value * self.width) // self.max_value
            sys.stdout.write("\r{}: [{}{}] {}/{} ".format(self.label, '#'*progress, ' '*(self.width-progress), value, self.max_value))
            sys.stdout.flush()
            self.last_time_printed = now
            return True
        return False

def command_line_input(item : str, options : OrderedDict, single_choice = False):
    if not single_choice:
        if "a" in options: raise AssertionError("options should not include key 'a'")
        if "d" in options: raise AssertionError("options should not include key 'd'")
        if "c" in options: raise AssertionError("options should not include key 'c'")
    if len(options) == 0: raise AssertionError("options should not be empty.")
    longest_option_descriptions = []
    longest_option_descriptions.append(max([len(key) for key in options] + [4]) + 4)
    i = 0
    while True:
        longest = -1
        for key in options:
            if i < len(options[key]):
                longest = max(longest, len(options[key][i]))
        if longest >= 0:
            longest_option_descriptions.append(longest + 4)
        else:
            break
        i += 1
        
    selected_keys = []      
    while True:
        keys = []
        print("Select {}.".format(item))
        print("    Option" + " " * (longest_option_descriptions[0] - len("Option")) + "Description")
        print("----" + "-" * sum(longest_option_descriptions))
        for key in options:
            keys.append(key)
            description = ""
            for i in range(len(options[key])):
                description += "{}{}".format(options[key][i], " " * (longest_option_descriptions[i+1] - len(options[key][i])))
            print("{}{}{}".format("[X] " if key in selected_keys else "[ ] ", key + " " * (longest_option_descriptions[0] - len(key)), description))
        if not single_choice:
            keys.append("a")
            print("    {}Select all".format("a" + " " * (longest_option_descriptions[0] - 1)))
        if not single_choice and len(selected_keys) > 0:
            keys.append("c")
            print("    {}Clear selection".format("c" + " " * (longest_option_descriptions[0] - 1)))
            keys.append("d")
            print("    {}done".format("d" + " " * (longest_option_descriptions[0] - 1)))
        selection = input("Enter option: ")
        if selection in keys:            
            if selection in options:
                selected_keys.append(selection)
                if single_choice:
                    break
            elif selection == "a":
                selected_keys = keys[:len(options)]
                break
            elif selection == "d":
                break
            elif selection == "c":
                selected_keys = []
        else:
            print ("Invalid selection. Enter any of {} or press Ctrl+C to abort.".format(keys))  
    print("Selected {}: {}".format(item,selected_keys))
    return selected_keys
            
            

class Settings(object):
    def __init__(self):
        self.settings_filename = "settings.json"
        self.json_data = OrderedDict()
        if os.path.isfile(self.settings_filename):
            self.json_data = load_json(self.settings_filename)
        if self.set_defaults():
            self.save()

    def set_defaults(self):
        
        set_an_option = False
        if not "benchmarks-directory" in self.json_data:
            self.json_data["benchmarks-directory"] = os.path.realpath(os.path.join(sys.path[0], "qcomp/benchmarks/"))
            set_an_option = True
        if not "benchmark-selection" in self.json_data:
            self.json_data["benchmark-selection"] = os.path.realpath(os.path.join(sys.path[0], "internal/qcomp2019_benchmarks_compatible.csv"))
            set_an_option = True
        if not "logs-directory-name" in self.json_data:
            self.json_data["logs-directory-name"] = "logs/"
            set_an_option = True
        if not "results-file-scatter" in self.json_data:
            self.json_data["results-file-scatter"] = "scatter.csv"
            set_an_option = True
        if not "results-file-quantile" in self.json_data:
            self.json_data["results-file-quantile"] = "quantile.csv"
            set_an_option = True
        if not "results-dir-table" in self.json_data:
            self.json_data["results-dir-table"] = "table/"
            set_an_option = True
        if not "time-limit" in self.json_data:
            self.json_data["time-limit"] = 1800
            set_an_option = True
        if not "goal-precision" in self.json_data:
            self.json_data["goal-precision"] = 1E-3
            set_an_option = True
        if not "relative-precision" in self.json_data:
            self.json_data["relative-precision"] = True
            set_an_option = True
        if not "storm-binary-dir" in self.json_data:
            self.json_data["storm-binary-dir"] = os.path.expanduser("~/storm/build/bin/")
            set_an_option = True
        if not "storm-configurations" in self.json_data:
            auto = Configuration(id = "automatic", note = "Automatic selection", command = "--engine automatic --sylvan:maxmem 4096 --sylvan:threads 4")
            sparse = Configuration(id = "sparse", note = "Sparse engine (default)", command = "--engine sparse")
            hybrid = Configuration(id = "hybrid", note = "Hybrid engine with Sylvan using 4GB memory", command = "--engine hybrid --sylvan:maxmem 4096 --sylvan:threads 4")
            dd = Configuration(id = "dd", note = "Symbolic engine with Sylvan using 4GB memory", command = "--engine dd --sylvan:maxmem 4096 --sylvan:threads 4")
            sound = Configuration(id = "sound", note = "Sparse engine with sound model checking", command = "--sound")
            exact = Configuration(id = "exact", note = "Sparse engine with exact model checking", command = "--exact")
            ddbisim = Configuration(id = "ddbisim", note = "Symbolic bisimulation with sparse quotient and sylvan using 4GB memory", command = "--engine dd-to-sparse --bisimulation --sylvan:maxmem 4096 --sylvan:threads 4")
            #dft = Configuration(id = "dft", note = "Storm-dft with symmetry reduction", command = "--symmetryreduction")
            self.json_data["storm-configurations"] = [auto.to_json(), sparse.to_json(), hybrid.to_json(), dd.to_json(), sound.to_json(), exact.to_json(), ddbisim.to_json()]
            set_an_option = True
        if not "filtered-paths" in self.json_data:
            self.json_data["filtered-paths"] = [os.path.realpath(sys.path[0]) + "/", os.path.expanduser("~") + "/"]
            set_an_option = True
        return set_an_option

    def benchmark_dir(self):
        """ Retrieves the directory where the qcomp benchmarks lie. """
        return self.json_data["benchmarks-directory"]
        
    def benchmark_selection(self):
        """ Retrieves the file which contains the benchmark selection. """
        return self.json_data["benchmark-selection"]

    def logs_dir(self):
        """ Retrieves the directory in which the tool logs are stored. """
        return self.json_data["logs-directory-name"]
    
    def input_logs_dir(self):
        """ Asks the user to enter a directory in which the tool logs are stored."""
        response = input("Enter a directory in which the tool logs are stored [{}]: ".format(self.logs_dir()))
        if response != "":
            self.json_data["logs-directory-name"] = response
            self.save()
    
    def results_file_scatter(self):
        """ Retrieves the filename to which the tool execution results for scatter plots are stored. """
        return self.json_data["results-file-scatter"]

    def results_file_quantile(self):
        """ Retrieves the filename to which the tool execution results for quantile plots are stored. """
        return self.json_data["results-file-quantile"]

    def results_dir_table(self):
        """ Retrieves the directory to which the tool execution result table is stored. """
        return self.json_data["results-dir-table"]

    def time_limit(self):
        """ Retrieves the time limit for tool executions (in seconds). """
        return int(self.json_data["time-limit"])

    def input_time_limit(self):
        """ Asks the user to enter a time limit."""
        while True:
            response = input("Enter a time limit (in seconds) after which executions are aborted [{}]: ".format(self.time_limit()))
            if response == "": break
            if is_number(response):
                self.json_data["time-limit"] = int(response)
                self.save()
                break
            else:
                print("Time limit should be a number.")
    
    def goal_precision(self):
        """ Retrieves the precision the tools have to achieved for numerical results. """
        return Fraction(self.json_data["goal-precision"])

    def is_relative_precision(self):
        """ Retrieves whether the precision is with respect to the relative error. """
        return bool(self.json_data["relative-precision"])
    
    def storm_binary_dir(self):
        """ Retrieves the directory in which the storm binaries are."""
        return self.json_data["storm-binary-dir"]
    
    def input_storm_binary_dir(self):
        """ Asks the user to enter a path to the storm binaries."""
        response = input("Enter path to storm binary directory [{}]: ".format(self.storm_binary_dir()))
        if response != "":
            self.json_data["storm-binary-dir"] = response
            self.save()
    
    def storm_configurations(self):
        """ Retrieves the known storm configurations """
        return [ Configuration(js) for js in self.json_data["storm-configurations"] ]

    def input_storm_configurations(self):
        """ Asks the user to edit existing or enter a new configuration """
        existing_configurations = self.storm_configurations()   
        while True:
            if len(existing_configurations) > 0 :
                print("The following configurations are available.")
                longest_config_note = max([len(config.note) for config in existing_configurations] + [len("Description")]) + 4
                longest_config_id = max([len(config.identifier) for config in existing_configurations] + [len("Identifier")]) + 4
                print("\tIdentifier{}Description{}Command".format(" " * (longest_config_id - len("Identifier")), " " * (longest_config_id - len("Identifier")) ))
                print("\t" + "-" * (longest_config_note + longest_config_id + len("Command ")))
                for config in existing_configurations:
                    print("\t{}{}{}{}{}".format(config.identifier, " " * (longest_config_id - len(config.identifier)), config.note, " " * (longest_config_note - len(config.note)), "(default)" if config.command == "" else config.command))
            id_found = False
            while not id_found:
                id_found = True
                print("To edit an existing or enter a new configuration, please enter an existing or a new configuration identifier. To select configurations for benchmarking, just hit Return.")
                id = input("Configuration Identifier: ")
                if id == "": break
                if "." in id:
                    print("Invalid: Id should not contain '.'.")
                    id_found = False
                if id in ["a", "c", "d"]:
                    print("Invalid: Id should not be any of 'a', 'b', or 'c'.")
                    id_found = False
                if not is_valid_filename(id):
                    print("Invalid: Id is not a valid filename.")
                    id_found = False
            if id == "": break
            conf = Configuration(id = id)
            existing_configurations = self.storm_configurations()
            i = 0
            for existing_conf in existing_configurations:
                if existing_conf.identifier == conf.identifier:
                    conf = Configuration(existing_conf.to_json())
                    print( "Found existing configuration.")
                    break
                i += 1
            note = input("Enter a description for the configuration{}: ".format(" [{}]".format(conf.note) if conf.note is not None else ""))
            if note != "" or conf.note is None:
                conf.note = note
            default_setting = ""
            if conf.command is not None:
                default_setting = " [{}]".format("default" if conf.command == "" else conf.command)
            command = input("Enter command line arguments for Storm that invoke the configuration ('default' means no additional arguments){}: ".format(default_setting))
            if command != "" or conf.command is None:
                if command == "default":
                    conf.command = ""
                else:
                    conf.command = command
            if i < len(existing_configurations):
                existing_configurations[i] = conf
            else:
                existing_configurations.append(conf)
            self.json_data["storm-configurations"] = [c.to_json() for c in existing_configurations]
            self.save()

    def filtered_paths(self):
        """ returns a list of paths (e.g. home directory) that should be filtered from commands in logfiles """
        return self.json_data["filtered-paths"]
        
    def save(self):
        save_json(self.json_data, self.settings_filename)
        print("Settings saved to {}.".format(os.path.realpath(self.settings_filename)))
