from internal.benchmark import *
from internal.utility import *
from internal import storm
from internal.invocation import *

import traceback



def create_invocations(settings):
    # Get and test directory of storm binary
    settings.input_storm_binary_dir()
    test_result = storm.test_installation(settings)
    if test_result is not None:
        print(test_result)
        input("Press Return to continue or CTRL+C to abort.")
    
    # Get and test configurations
    settings.input_storm_configurations()
    existing_configurations = settings.storm_configurations()
    for config in existing_configurations:
        test_result = storm.test_installation(settings, config)
        if test_result is not None:
            print("Error when testing configuration '{}':".format(config.identifier))
            print(test_result)
            input("Press Return to continue or CTRL+C to abort.")
        
        
    # configuration selection
    selected_config_identifiers = command_line_input("configurations", OrderedDict([(config.identifier, [config.note, "default" if config.command == "" else config.command]) for config in existing_configurations]))
    selected_configurations = []
    for config in existing_configurations:
        if config.identifier in selected_config_identifiers:
            selected_configurations.append(config)
    print("Selected {} configurations.".format(len(selected_configurations)))
        
    # Loading benchmarks
    qcomp_benchmarks_csv = load_csv(os.path.join(sys.path[0], settings.benchmark_selection()))
    qcomp_benchmarks = OrderedDict([("dtmc",[]),("ctmc",[]),("mdp",[]),("ma",[]),("pta",[])])
    for benchmark_csv in qcomp_benchmarks_csv:
        benchmark_id = "{}.{}.{}".format(benchmark_csv[0], benchmark_csv[3], benchmark_csv[4])
        benchmark = get_benchmark_from_id(settings, benchmark_id)
        qcomp_benchmarks[benchmark.get_model_type()].append(benchmark)    
    print("Loaded {} benchmarks from selection '{}'".format(len(qcomp_benchmarks_csv), settings.benchmark_selection()))
    
    # model type selection
    model_types = OrderedDict()
    model_types["dtmc"] = ["Discrete Time Markov Chains", "({} benchmarks)".format(len(qcomp_benchmarks["dtmc"]))]
    model_types["ctmc"] = ["Continuous Time Markov Chains", "({} benchmarks)".format(len(qcomp_benchmarks["ctmc"]))]
    model_types["mdp"] = ["Markov Decision Processes", "({} benchmarks + {} PTAs via moconv)".format(len(qcomp_benchmarks["mdp"]), len(qcomp_benchmarks["pta"]))]
    model_types["ma"] = ["Markov Automata", "({} benchmarks)".format(len(qcomp_benchmarks["ma"]))]
    qcomp_benchmarks["mdp"] += qcomp_benchmarks["pta"]
    selected_model_types = command_line_input("models", model_types)
    
    # property type selection
    selected_benchmarks_per_property = OrderedDict()
    for model_type in selected_model_types:
        for benchmark in qcomp_benchmarks[model_type]:
            if benchmark.get_property_type() in selected_benchmarks_per_property:
                selected_benchmarks_per_property[benchmark.get_property_type()].append(benchmark)
            else:
                selected_benchmarks_per_property[benchmark.get_property_type()] = [benchmark]
    property_types = OrderedDict()
    for property_type in selected_benchmarks_per_property:
        property_types[str(len(property_types) + 1)] = [property_type, " ({} benchmarks)".format(len(selected_benchmarks_per_property[property_type]))]
    selected_property_types = command_line_input("properties", property_types)
    selected_benchmarks = []
    for property_type_number in selected_property_types:
        property_type_str = property_types[property_type_number][0] # e.g. 'mdp (7)'
        for benchmark in selected_benchmarks_per_property[property_type_str]:
            selected_benchmarks.append(benchmark)       
    num_invocations = len(selected_benchmarks) * len(selected_configurations)
    print("Selected {} benchmarks and {} configurations yielding {} invocations in total.".format(len(selected_benchmarks), len(selected_configurations), num_invocations))
    
    settings.input_time_limit()
    settings.input_logs_dir()
    
    # Creating invocations
    invocations = []
    progressbar = Progressbar(num_invocations, "Generating invocations")
    i = 0
    unsupported = []
    for benchmark in selected_benchmarks:
        for configuration in selected_configurations:
            i += 1
            progressbar.print_progress(i)
            invocation = storm.get_invocation(settings, benchmark, configuration)
            if len(invocation.commands) == 0:
                unsupported.append(invocation.get_identifier() + ": " + invocation.note)
            else:
                invocation.time_limit = settings.time_limit()
                invocations.append(invocation) 
    print("")
    if len(unsupported) > 0:
        print("{} invocations are not supported:".format(len(unsupported)))
        for inv in unsupported:
            print("\t" + inv)
    return invocations

def check_invocations(settings, invocations):
    ensure_directory(settings.logs_dir())
    invocation_number = 0
    if len(invocations) > 1:
        progressbar = Progressbar(len(invocations), "Checking invocations")
    else:
        sys.stdout.write("Checking invocation ... ")
        sys.stdout.flush()
    invocation_identifiers = set()
    for invocation in invocations:
        invocation_number = invocation_number + 1
        if len(invocations) > 1:
            progressbar.print_progress(invocation_number)
        try:
            # check whether there are no commands
            if len(invocation.commands) == 0:
                continue
            benchmark = get_benchmark_from_id(settings, invocation.benchmark_id)
            # ensure that the actual benchmark files exist
            for filename in benchmark.get_all_filenames():
                if not os.path.isfile(os.path.join(benchmark.get_directory(), filename)):
                    raise AssertionError(
                        "The file '{}' does not exist.".format(os.path.join(benchmark.get_directory(), filename)))
            # ensure that the invocation identifier (consisting of benchmark and configuration id) can be a filename and are unique
            if not is_valid_filename(invocation.get_identifier(), "/"):
                raise AssertionError("Invocation identifier '{}' is either not a valid filename or contains a '.'.".format(invocation.get_identifier()))
            if invocation.get_identifier() in invocation_identifiers:
                raise AssertionError("Invocation identifier '{}' already exists.".format(invocation.get_identifier()))
            invocation_identifiers.add(invocation.get_identifier())
        except Exception:
            print("Error when checking invocation #{}: {}".format(invocation_number - 1, invocation.get_identifier()))
            traceback.print_exc()
    print("")
    

def run_invocations(settings, invocations):
    invocation_number = 0
    if len(invocations) > 1:
        progressbar = Progressbar(len(invocations), "Executing invocations")
    else:
        sys.stdout.write("Executing invocation ... ")
        sys.stdout.flush()
    try:
        for invocation in invocations:
            invocation_number = invocation_number + 1
            benchmark = get_benchmark_from_id(settings, invocation.benchmark_id)
            if len(invocations) > 1:
                progressbar.print_progress(invocation_number)
            # execute the invocation
            notes = []
            execution = invocation.execute()
            tool_result = execution.to_json()
            try:
                result = storm.get_result(benchmark, execution)
            except Exception as e:
                print("ERROR while getting result for invocation #{}: {}".format(invocation_number - 1, invocation.get_identifier()))
                traceback.print_exc()
                result = None
            if result is not None:
                tool_result["result"] = str(result)  # convert to str to not loose precision
                result = try_to_bool_or_number(result)
                if benchmark.has_reference_result():
                    tool_result["result-correct"] = is_result_correct(settings, benchmark.get_reference_result(), result)
                    if is_number(result) and is_number_or_interval(benchmark.get_reference_result()):
                        tool_result["absolute-error"] = try_to_float(get_absolute_error(benchmark.get_reference_result(), result))
                        tool_result["relative-error"] = try_to_float(get_relative_error(benchmark.get_reference_result(), result))
                        if not tool_result["result-correct"]:
                            # Prepare a message
                            if settings.is_relative_precision():
                                error_kind = "a relative"
                                error_value = tool_result["relative-error"]
                            else:
                                error_kind = "an absolute"
                                error_value = tool_result["absolute-error"]
                            if is_number(result) and not (isinstance(result,float) or isinstance(result,int)):
                                tool_result_str = "'{}' (approx. {})".format(result, try_to_float(result))
                            else:
                                tool_result_str = "'{}'".format(result)
                            if is_number(benchmark.get_reference_result()) and not (isinstance(benchmark.get_reference_result(), float) or isinstance(benchmark.get_reference_result(), int)):
                                ref_result_str = "'{}' (approx. {})".format(benchmark.get_reference_result(), try_to_float(benchmark.get_reference_result()))
                            elif (isinstance(benchmark.get_reference_result(), dict) or isinstance(benchmark.get_reference_result(), OrderedDict)) and "lower" in benchmark.get_reference_result() and "upper" in benchmark.get_reference_result():
                                ref_result_str = "[{},{}]".format(try_to_float(benchmark.get_reference_result()["lower"], benchmark.get_reference_result()["upper"]))
                            else:
                                ref_result_str = "'{}'".format(benchmark.get_reference_result())
                            notes.append("The tool result {} is tagged as incorrect. The reference result is {} which means {} error of '{}' which is larger than the goal precision '{}'.".format(tool_result_str, ref_result_str, error_kind, error_value, try_to_float(settings.goal_precision())))
                    elif not tool_result["result-correct"]:
                        notes.append("Result '{}' is tagged as incorrect because it is different from the reference result '{}'.".format(result, benchmark.get_reference_result()))
                else:
                    notes.append("Correctness of result is not checked because no reference result is available.")
            elif not execution.timeout and not execution.error:
                notes.append("Unable to obtain tool result.")
                tool_result["execution-error"] = True
            tool_result["notes"] = notes
            logfile_name = invocation.get_identifier() + ".log"
            tool_result["log"] = logfile_name
            # save logfile
            with open(os.path.join(settings.logs_dir(), logfile_name), 'w', encoding="utf-8") as logfile:
                logfile.write(execution.concatenate_logs())
                if len(notes) > 0:
                    logfile.write("\n" + "#"*30 + " Notes " + "#"*30 + "\n")
                for note in notes:
                    logfile.write(note + "\n")
            # save tool results in json format
            save_json(tool_result, os.path.join(settings.logs_dir(), invocation.get_identifier() + ".json"))
    except KeyboardInterrupt as e:
        print("\nInterrupt while processing invocation #{}: {}".format(invocation_number - 1, invocation.get_identifier()))
    except Exception:
        print("ERROR while processing invocation #{}: {}".format(invocation_number - 1, invocation.get_identifier()))
        traceback.print_exc()
    
if __name__ == "__main__":
    print("Storm benchmarking tool.")
    print("This script selects and executes benchmarks.")
    print("Usages:")
    print("python3 {}                 Creates and executes a set of invocations.".format(sys.argv[0]))
    print("python3 {} <filename>      Executes benchmarks from a previously created invocations file located at <filename>.".format(sys.argv[0]))
    print("python3 {} <filename> <i>  Executes the <i>th invocation (0 based) from a previously created invocations file located at <filename>.".format(sys.argv[0]))
    print("")
    if (len(sys.argv) == 2 and sys.argv[1] in ["-h", "-help", "--help"]) or len(sys.argv) > 3:
        exit(1)
    
    settings = Settings()
    
    if len(sys.argv) == 1:
        input("No invocations file loaded. Press Return to create one now or CTRL+C to abort.")
        invocations = create_invocations(settings)
        while True:
            response = input("Enter a filename to store the invocations for later usage or press Return to continue: ")
            if response == "": break
            if (is_valid_filename(response)):
                if os.path.isfile(response): 
                    if input("File {} exists. Overwrite? (type 'y' or 'n'): ".format(response)) != "y":
                        continue
                invocations_json = [inv.to_json() for inv in invocations]
                save_json(invocations_json, response)
                print("Saved invocations to file '{}'.".format(response))
                print("To load these invocations at a later time, you may run\n\tpython3 {} {}".format(sys.argv[0], response))
                break
            else:
                print("Invalid file name {}".format(response))
    else:
        if not os.path.isfile(sys.argv[1]):
            raise AssertionError("Invocations file {} does not exist".format(sys.argv[1]))
        invocations_json = load_json(sys.argv[1])
        invocations = [Invocation(inv) for inv in invocations_json]
        print("Loaded {} invocations.".format(len(invocations)))
        if len(sys.argv) == 3:
            if not is_number(sys.argv[2]): raise AssertionError("Expected a number for second argument but got '{}' instead.".format(sys.argv[2]))
            selected_index = int(sys.argv[2])
            if selected_index not in range(0,len(invocations)): raise AssertionError("Second argument is out of range: got '{}' but index has to be at least 0 at less than {}".format(selected_index, len(invocations)))
            invocations = [invocations[selected_index]]    
            print("Selected invocation #{}: {}".format(selected_index, invocations[0].get_identifier()))
            
    check_invocations(settings, invocations)
    run_invocations(settings, invocations)
    
    
    

    
    